
/*****************************************************************************
 *
 * Tree.C
 *
 *****************************************************************************/

/*****************************************************************************
 *
 * Headers
 *
 *****************************************************************************/

#include <stdio.h>

#include "intrinsic.h"
#include "mem.h"
#include "tree.h"

/*****************************************************************************
 *
 * Local Constants and Macros
 *
 *****************************************************************************/

#define LIB$_KEYALRINS   1409057

#if defined(VMS) || defined(__GNUC__)
#define _PROTOTYPES_
#endif

/*****************************************************************************
 *
 * Local Declarations
 *
 *****************************************************************************/

static Tree mms_xfree_tree = NULL;
static Tree mms_global_root;
static Tree mms_tree_node;
static int  mms_tree_fcn1,mms_tree_fcn2;

/*****************************************************************************
 *
 * Local Prototypes
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int  mms_i_tree_alloc          (int key, int *ptr);
static int  mms_internal_traverse     (Tree node);
static int  mms_internal_traverse2    (Tree node);
static int  mms_tree_default_compare  (String k, Tree t);
static int  mms_tree_default_key      (String k, Tree t);
static int  mms_tree_idefault_compare (int k, Tree t);
static int  mms_tree_idefault_key     (int k, Tree t);
static int  xmms_tree_balance1        (Tree *root, int *h);
static int  xmms_tree_balance2        (Tree *root, int *h);
static int  xmms_tree_insert          (Tree *root, String key, int *ins_flag,
                                       IPFN comp_fcn, IPFN alloc_fcn, 
                                       Tree *rptr, Pointer data);
#else
static int  mms_i_tree_alloc          ();
static int  mms_internal_traverse     ();
static int  mms_internal_traverse2    ();
static int  mms_tree_default_compare  ();
static int  mms_tree_default_key      ();
static int  mms_tree_idefault_compare ();
static int  mms_tree_idefault_key     ();
static int  xmms_tree_balance1        ();
static int  xmms_tree_balance2        ();
static int  xmms_tree_insert          ();
#endif

/*****************************************************************************
 *
 * DeallocTreeNode
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
extern int DeallocTreeNode (Tree ptr)
#else
extern int DeallocTreeNode (ptr)

  Tree ptr;
#endif

{
  ptr->left      = mms_xfree_tree;
  mms_xfree_tree = ptr;

  return(Success);
}

/*****************************************************************************
 *
 * DeallocTree
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
extern Status DeallocTree (Tree ptr)
#else
extern Status DeallocTree (ptr)

  Tree ptr;
#endif

{
  /* TBD */

  /* done */
  return (Failure);
}

/*****************************************************************************
 *
 * TreeInsert
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
extern Tree TreeInsert (int *inskey, Tree *tree, String key, Pointer data,
                 IPFN comp_fcn, IPFN key_fcn)
#else
extern Tree TreeInsert (inskey, tree, key, data, comp_fcn, key_fcn)

  int     *inskey;
  Tree    *tree;
  String   key;
  Pointer  data;
  IPFN     comp_fcn;
  IPFN     key_fcn;
#endif

{
  int         i,j,k;
  int         mms_tree_default_compare(),mms_tree_default_key();
  int         mms_tree_idefault_compare(),mms_tree_idefault_key();
  int         (*cfcn)(),(*kfcn)();
  Tree        t1;
  int         mms_i_tree_alloc();

  if (((int) comp_fcn) == 1) cfcn = mms_tree_idefault_compare;
  else if ((int) comp_fcn)   cfcn = comp_fcn;
  else                       cfcn = mms_tree_default_compare;
  if (((int) key_fcn) == 1)  kfcn = mms_tree_idefault_key;
  else if ((int) key_fcn)    kfcn = key_fcn;
  else                        kfcn = mms_tree_default_key;
  j = *inskey;
  i = xmms_tree_insert (tree,key,&j,cfcn,mms_i_tree_alloc,&t1,0);
  *inskey = 0;
  if (i == LIB$_KEYALRINS && !t1->extra) {
    t1->extra = 1;
    t1->data = data;
  }
  else if (i == LIB$_KEYALRINS) *inskey = 1;
  else {
    t1->extra = 1;
    t1->data = data;
    (*kfcn)(key,t1);
  }

  return(t1);
}

/*****************************************************************************
 *
 * mms_tree_default_compare
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int mms_tree_default_compare (String k, Tree t)
#else
static int mms_tree_default_compare (k, t)

  String k;
  Tree   t;
#endif

{
  return (strcmp(k,TreeKey(t)));
}

/*****************************************************************************
 *
 * mms_tree_default_key
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int mms_tree_default_key (String k, Tree t)
#else
static int mms_tree_default_key (k, t)

  String k;
  Tree   t;
#endif

{
  String s;

  s = (String) MemAlloc (strlen(k) + 1, NoInitialize);
  strcpy (s, k);
  TreeKey(t) = s;
  return(Success);
}

/*****************************************************************************
 *
 * mms_tree_idefault_compare
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int mms_tree_idefault_compare (int k, Tree t)
#else
static int mms_tree_idefault_compare (k, t)

  int  k;
  Tree t;
#endif

{
  return (k - ((int) TreeKey(t)));
}

/*****************************************************************************
 *
 * mms_tree_idefault_key
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int mms_tree_idefault_key (int k, Tree t)
#else
static int mms_tree_idefault_key (k, t)

  int  k;
  Tree t;
#endif

{
  TreeKey(t) = (String) k;
  return(Success);
}

/*****************************************************************************
 *
 * mms_i_tree_alloc
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int mms_i_tree_alloc (int key, int *ptr)
#else
static int mms_i_tree_alloc (key, ptr)

  int  key;
  int *ptr;
#endif

{
  int  i,j,k;
  Tree t1;

  if (mms_xfree_tree) {
    *ptr           = (int) mms_xfree_tree;
    mms_xfree_tree = mms_xfree_tree->left;
  }
  else {
    *ptr = (int) MemAlloc (sizeof(_Tree), NoInitialize);
  }

  return(Success);
}

/*****************************************************************************
 *
 * TreeTraverse
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
extern Status TreeTraverse (Tree tree, IPFN fcn)
#else
extern Status TreeTraverse (tree, fcn)

  Tree tree;
  IPFN fcn;
#endif

{
  int         stack[24][2];
  int         stack_size;
  Tree        ptr,ptr2;
  int         i,j;

  ptr        = tree;
  stack_size = 0;

  if (!ptr) return(1);

  stack[stack_size][0] = (int) (ptr);
  stack[stack_size][1] = 0;
  stack_size += 1;

  while (stack_size) {

    i = stack_size - 1;

    ptr = (Tree) stack[i][0];

    if (stack[i][1] == 0) {
      if (TreeLeft(ptr)) {
        stack[stack_size][0] = (int) (TreeLeft(ptr));
        stack[stack_size][1] = 0;
        stack_size += 1;
      }
      stack[i][1] = 1;
    }
    else {
      if (fcn) {
        j = (*fcn) (ptr);
        if (!(j & 1)) break;
      }
      stack_size -= 1;
      if (TreeRight(ptr)) {
        stack[stack_size][0] = (int) (TreeRight(ptr));
        stack[stack_size][1] = 0;
        stack_size += 1;
      }
    }
  }

  return(Success);
}

/*****************************************************************************
 *
 * mms_internal_traverse
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int mms_internal_traverse (Tree tree)
#else
static int mms_internal_traverse (tree)

  Tree tree;
#endif

{
  printf ("`%s' %d\n",TreeKey(tree),TreeData(tree));

  return(Success);
}

/*****************************************************************************
 *
 * mms_internal_traverse2
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int mms_internal_traverse2 (Tree tree)
#else
static int mms_internal_traverse2 (tree)

  Tree tree;
#endif

{
  printf ("%s\n",TreeKey(tree));

  return(Success);
}

/*****************************************************************************
 *
 * TreeFind
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
extern Tree TreeFind (Tree tree, String key, IPFN comp_fcn)
#else
extern Tree TreeFind (tree, key, comp_fcn)

  Tree   tree;
  String key;
  IPFN   comp_fcn;
#endif

{
  Tree ptr;
  int  i;

  ptr = tree;

  while (ptr) {
    if (!comp_fcn)                  i = strcmp(key,TreeKey(ptr));
    else if (((int) comp_fcn) == 1) i = ((int)key) - ((int)(TreeKey(ptr)));
    else                            i = (*comp_fcn)(key,(ptr));
    if (!i) break;
    else if (i < 0) ptr = TreeLeft(ptr);
    else            ptr = TreeRight(ptr);
  }

  return(ptr);
}

/*****************************************************************************
 *
 * TreeReplaceData
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
extern int TreeReplaceData (Tree tptr,Pointer newdata, IPFN fcn)
#else
extern int TreeReplaceData (tptr, newdata, fcn)

  Tree    tptr;
  Pointer newdata;
  IPFN    fcn;
#endif

{
  int  i,j,k;
  Tree tree;

  tree = tptr;

  if (fcn) (*fcn)(tree->data);
  tree->data = newdata;

  return(Success);
}

/*****************************************************************************
 *
 * xmms_tree_find
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int xmms_tree_find (Tree ptr,String key,IPFN comp_fcn)
#else
static int xmms_tree_find (ptr, key, comp_fcn)

  Tree   ptr;
  String key;
  IPFN   comp_fcn;
#endif


{
  int i;

  while (ptr) {
    i = (*comp_fcn)(key,(ptr));
    if (!i) break;
    else if (i < 0) ptr = TreeLeft(ptr);
    else            ptr = TreeRight(ptr);
  }

  return((int) ptr);
}

/*****************************************************************************
 *
 * xmms_tree_insert
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int xmms_tree_insert (Tree *root, String key, int *ins_flag,
                             IPFN comp_fcn, IPFN alloc_fcn, Tree *rptr,
                             Pointer data)
#else
static int xmms_tree_insert (root, key, ins_flag, comp_fcn, alloc_fcn,
                             rptr, data)

  Tree   *root;
  String  key;
  int    *ins_flag;
  IPFN    comp_fcn;
  IPFN    alloc_fcn;
  Tree   *rptr;
  Pointer data;
#endif

{
  int         i,j,k;
  static int  stack[256][2];
  static int  stack_ptr;
  int         h;
  int         dup_flag;
  int         inserted;
  int         return_ptr;
  int         return_value;
  Tree       *tptr;
  Tree        ptr,ptr2,ptr3;

  return_value          = 1;
  h                     = 0;
  dup_flag              = *ins_flag;
  stack_ptr             = 0;
  inserted              = 0;
  return_ptr            = 0;
  stack[stack_ptr][0]   = (int) root;
  stack[stack_ptr++][1] = 0;

  while (stack_ptr) {

    stack_ptr--;
    tptr = (Tree *) stack[stack_ptr][0];
    ptr  = *tptr;

    if (stack[stack_ptr][1] == 0) {
      if (ptr) i = (*comp_fcn) (key,(ptr));
      else     i = 0;
      if (!ptr) {
        (*alloc_fcn) (key,&ptr2);
        ptr2->left    = ptr2->right = NULL;
        ptr2->data    = data;
        ptr2->extra   = 0;
        ptr2->balance = 0;
        *tptr         = ptr2;
        return_ptr    = (int) ptr2;
        h             = 1;
        inserted      = 1;
      }
      else if (!i && !dup_flag) {
        inserted     = 0;
        return_ptr   = (int) ptr;
        return_value = LIB$_KEYALRINS;
      }
      else if (i < 1) {
        stack[stack_ptr][0] = (int) tptr;
        stack[stack_ptr][1] = 1;
        stack_ptr++;
        stack[stack_ptr][0] = (int) (&(TreeLeft(ptr)));
        stack[stack_ptr][1] = 0;
        stack_ptr++;
      }
      else {
        stack[stack_ptr][0] = (int) tptr;
        stack[stack_ptr][1] = 2;
        stack_ptr++;
        stack[stack_ptr][0] = (int) (&(TreeRight(ptr)));
        stack[stack_ptr][1] = 0;
        stack_ptr++;
      }
    }
  
    else if (stack[stack_ptr][1] == 1) {
      if (h) {
        if (TreeBalance(ptr) == 1) {
          ptr->balance = 0;
          h = 0;
        }
        else if (TreeBalance(ptr) == 0) {
          ptr->balance = -1;
        }
        else if (TreeBalance(ptr) == -1) {
          ptr2 = ptr->left;
          if (TreeBalance(ptr2) == -1) { /* single LL rotation */
            TreeLeft(ptr)   = TreeRight(ptr2);
            TreeRight(ptr2) = ptr;
            TreeBalance(ptr)    = 0;
            ptr              = ptr2;
          }
          else { /* double LR rotation */
            ptr3             = TreeRight(ptr2);
            TreeRight(ptr2) = TreeLeft(ptr3);
            TreeLeft(ptr3)  = ptr2;
            TreeLeft(ptr)   = TreeRight(ptr3);
            TreeRight(ptr3) = ptr;
            if (ptr3->balance == -1) ptr->balance  = 1;
            else                          ptr->balance  = 0;
            if (ptr3->balance ==  1) ptr2->balance = -1;
            else                          ptr2->balance =  0;
            ptr = ptr3;
          }
          ptr->balance = 0;
          h     = 0;
          *tptr = ptr;
        }
      }
    }
  
    else if (stack[stack_ptr][1] == 2) {
      if (h) {
        if (ptr->balance == -1) {
          ptr->balance = 0;
          h                 = 0;
        }
        else if (ptr->balance == 0) {
          ptr->balance = 1;
        }
        else if (ptr->balance == 1) {
          ptr2 = ptr->right;
          if (ptr2->balance == 1) { /* single RR rotation */
            ptr->right   = ptr2->left;
            ptr2->left   = ptr;
            ptr->balance = 0;
            ptr          = ptr2;
          }
          else { /* double RL rotation */
            ptr3        = ptr2->left;
            ptr2->left  = ptr3->right;
            ptr3->right = ptr2;
            ptr->right  = ptr3->left;
            ptr3->left  = ptr;
            if (ptr3->balance ==  1) ptr->balance  = -1;
            else                     ptr->balance  =  0;
            if (ptr3->balance == -1) ptr2->balance = 1;
            else                     ptr2->balance = 0;
            ptr = ptr3;
          }
          ptr->balance = 0;
          h            = 0;
          *tptr        = ptr;
        }
      }
    }
  }

  *ins_flag = inserted;
  *rptr     = (Tree) return_ptr;
  return(return_value);
}

/*****************************************************************************
 *
 * xmms_tree_delete1
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int xmms_tree_delete1 (Tree *root, String key, int *h, 
                              IPFN comp_fcn, IPFN key_fcn, IPFN free_fcn)
#else
static int xmms_tree_delete1 (root, key, h, comp_fcn, key_fcn, free_fcn)

  Tree   *root;
  String  key;
  int    *h;
  IPFN    comp_fcn;
  IPFN    key_fcn;
  IPFN    free_fcn;
#endif

{
  int  i,j,k;
  Tree q,p1,p2,p;
  int  b1,b2;

  if (!root) {
    *h = 0;
    return(0);
  }

  p = *root;
  if (!p) {
    *h = 0;
    return(0);
  }

  i = (*comp_fcn) (key,p);
  if (i < 0) {
    i = xmms_tree_delete1 (&(p->left),key,h,comp_fcn,key_fcn,free_fcn);
    if (*h) xmms_tree_balance1 (root,h);
    return(i);
  }
  else if (i > 0) {
    i = xmms_tree_delete1 (&(p->right),key,h,comp_fcn,key_fcn,free_fcn);
    if (*h) xmms_tree_balance2 (root,h);
    return(i);
  }
  else {

    q = p;
    if (!(q->right)) {
      *root = q->left;
      *h = 1;
      if (free_fcn) (*free_fcn) (q->data);
      DeallocTreeNode (q);
    }
    else if (!(q->left)) {
      *root = q->right;
      *h = 1;
      if (free_fcn) (*free_fcn) (q->data);
      DeallocTreeNode (q);
    }
    else {
      xmms_tree_del (&q,&(q->left),h,comp_fcn,key_fcn,free_fcn);
      *root = q;
      if (*h) xmms_tree_balance1 (root,h);
    }
    return(1);
  }
}

/*****************************************************************************
 *
 * xmms_tree_del
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
xmms_tree_del (Tree *root, Tree *root_r, int *h, IPFN comp_fcn, 
               IPFN key_fcn, IPFN free_fcn)
#else
xmms_tree_del (root, root_r, h, comp_fcn, key_fcn, free_fcn)

  Tree  *root;
  Tree  *root_r;
  int   *h;
  IPFN   comp_fcn;
  IPFN   key_fcn;
  IPFN   free_fcn;
#endif

{
  Tree r;

  r = *root_r;
  if (r->right) {
    xmms_tree_del (root,&(r->right),h, comp_fcn, key_fcn, free_fcn);
    if (*h) xmms_tree_balance2 (root_r,h);
  }
  else {
    (*key_fcn) (r->key, (*root));
    (*root)->extra = r->extra;
    (*root)->data = r->data;
    *root_r = r->left;
    *h = 1;
    if (free_fcn) (*free_fcn) (r->data);
    DeallocTreeNode (r);
  }

  return(1);
}

/*****************************************************************************
 *
 * xmms_tree_balance1
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int xmms_tree_balance1 (Tree *root, int *h)
#else
static int xmms_tree_balance1 (root, h)

  Tree *root;
  int  *h;
#endif

{
  Tree p,p1,p2;
  int  b1,b2;

  p = *root;
  if (p->balance == -1) p->balance = 0;
  else if (p->balance == 0) {
    p->balance = 1;
    *h = 0;
  }
  else if (p->balance == 1) {
    p1 = p->right;
    b1 = p1->balance;
    if (b1 >= 0) { /* single RR rotation */
      p->right = p1->left;
      p1->left = p;
      if (!b1) {
        p->balance = 1;
        p1->balance = -1;
        *h = 0;
      }
      else {
        p->balance = 0;
        p1->balance = 0;
      }
      p = p1;
      *root = p;
    }
    else { /* double RL rotation */
      p2 = p1->left;
      b2 = p2->balance;
      p1->left = p2->right;
      p2->right = p1;
      p->right = p2->left;
      p2->left = p;
      if (b2 == 1) p->balance = -1;
      else         p->balance = 0;
      if (b2 == -1) p1->balance = 1;
      else          p1->balance = 0;
      p = p2;
      p2->balance = 0;
      *root = p;
    }
  }

  return(1);
}

/*****************************************************************************
 *
 * xmms_tree_balance2
 *
 *****************************************************************************/

#ifdef _PROTOTYPES_
static int xmms_tree_balance2 (Tree *root, int *h)
#else
static int xmms_tree_balance2 (root, h)

  Tree  *root;
  int   *h;
#endif

{
  Tree p,p1,p2;
  int  b1,b2;

  p = *root;
  if (p->balance == 1) p->balance = 0;
  else if (p->balance == 0) {
    p->balance = -1;
    *h = 0;
  }
  else if (p->balance == -1) {
    p1 = p->left;
    b1 = p1->balance;
    if (b1 <= 0) { /* single LL rotation */
      p->left = p1->right;
      p1->right = p;
      if (b1 == 0) {
        p->balance = -1;
        p1->balance = 1;
        *h = 0;
      }
      else {
        p->balance = 0;
        p1->balance = 0;
      }
      p = p1;
      *root = p;
    }
    else { /* double LR rotation */
      p2 = p1->right;
      b2 = p2->balance;
      p1->right = p2->left;
      p2->left = p1;
      p->left = p2->right;
      p2->right = p;
      if (b2 == -1) p->balance = 1;
      else          p->balance = 0;
      if (b2 == 1) p1->balance = -1;
      else         p1->balance = 0;
      p = p2;
      p2->balance = 0;
      *root = p;
    }
  }

  return(1);
}
