
/******************************************************************************
 *
 * Tree.H
 *
 ******************************************************************************/

typedef int  *TreeData;
typedef char *TreeKey;

typedef struct __Tree {
  struct __Tree *left;
  struct __Tree *right;
  short          balance;
  short          extra;
  TreeKey        key;
  TreeData       data;
} _Tree, *Tree, TreeStruct, *TreePtr, *TreeP, vtree_t, vtree_ptr_t;

#define StringTree          0
#define NumberTree          1
#define NumericTree         1

#define TreeLeft(item)         ((item)->left)
#define TreeRight(item)        ((item)->right)
#define TreeData(item)         ((item)->data)
#define TreeExtra(item)        ((item)->extra)
#define TreeKey(item)          ((item)->key)
#define TreeBalance(item)      ((item)->balance)

#define TreeStackSize 128

typedef struct {
  int   key[TreeStackSize];
  Tree  stack[TreeStackSize];
  int   size;
} TreeStack, TreeStackStruct, *TreeStackPtr, *TreeStackP, _TreeStack;

#define TraverseTree(root,xstack)                                        \
  for ((xstack).size = 1, (xstack).stack[0] = root, (xstack).key[0] = 0; \
       (xstack).size;                                                    \
       (xstack).size)

#define TraverseTreeNode(xstack,node) {                    \
  (node) = (xstack).stack[(xstack).size - 1];              \
  if ((node) == 0) {                                       \
    (xstack).size -= 1;                                    \
    continue; /* ignore node */                            \
  }                                                        \
  if ((xstack).key[(xstack).size - 1] == 0) {              \
    (xstack).stack[(xstack).size]   = TreeLeft(node);      \
    (xstack).key[(xstack).size]     = 0;                   \
    (xstack).key[(xstack).size - 1] = 1;                   \
    (xstack).size                 += 1;                    \
    continue;  /* visit left */                            \
  }                                                        \
  else if ((xstack).key[(xstack).size - 1] == 1) {         \
    (xstack).key[(xstack).size - 1] = 2;                   \
  }                                                        \
  else {                                                   \
    (xstack).key[(xstack).size - 1]   = 0;                 \
    (xstack).stack[(xstack).size - 1] = TreeRight(node);   \
    continue;  /* visit right */                           \
  }                                                        \
}

extern int    DeallocTreeNode (Tree ptr);
extern Status DeallocTree     (Tree ptr);
extern Tree   TreeInsert      (int *inskey, Tree *tree, String key, 
                               Pointer data, IPFN comp_fcn, IPFN key_fcn);
extern Status TreeTraverse    (Tree tree, IPFN fcn);
extern Tree   TreeFind        (Tree tree, String key, IPFN comp_fcn);
extern int    TreeReplaceData (Tree tptr,Pointer newdata, IPFN fcn);
